#include "dynfun.h"

// #define DEBUG_ASM

int DynamicFunction:: create(Cell funname, const Kind_t k, Cell eqFun, 
			     Cell defaultValue) {
  return (new DynamicFunction(funname, k, eqFun, defaultValue))->id();
}

int DynamicFunction::create(Cell funname, const Kind_t k, Cell eqFun, 
			    Cell ordFun, Cell defaultValue, Cell initList) {
  return (new DynamicFunction(funname, k, eqFun, ordFun, defaultValue,
			      initList))->id();
}


void DynamicFunction::initClass() {
  Globals_t &g = getGlobals();
  g.locationsNew = &g.locationMapNew[0];
  g.locationsOld = &g.locationMapOld[0];
}



DynamicFunction::DynamicFunction(Cell funname, const Kind_t k, Cell eqFun, 
				 Cell defaultValue) 
  : m_useCounter(0), m_kind(k), m_globals(getGlobals())
{ 
  m_id = getNewID();
  m_globals.flist[m_id]  = this;
#ifdef DEBUG_ASM
  printf("creating dynamic function %d: ",m_id);
  ERREXPR(funname);
  printf("\n");
#endif
  m_globals.initCounter++;
  m_eqFunction   = eqFun;
  m_ordFunction  = NIL;
  m_funname      = funname;
  m_dynFun       = Dynamic0;
  m_values       = NULL;
  m_initList     = NIL;
  m_defaultValue = defaultValue;
  m_defaultValue = eval_value(m_defaultValue);
  m_globals.initCounter--;
}

DynamicFunction::DynamicFunction(Cell funname, const Kind_t k, Cell eqFun, 
				 Cell ordFun, Cell defaultValue, 
				 Cell initList)
  : m_useCounter(0), m_kind(k), m_globals(getGlobals())
{
  Cell arg, value, next;
  m_id = getNewID();
  m_globals.flist[m_id] = this;
  m_useCounter++;
#ifdef DEBUG_ASM
  printf("creating dynamic function %d: ",m_id);
  ERREXPR(funname);
  printf("\n");
#endif
  m_globals.initCounter++;
  m_eqFunction   = eqFun;
  m_ordFunction  = ordFun;
  m_funname      = funname;
  m_dynFun       = Dynamic1;
  m_initList     = initList;
  m_values       = new Values_t(LessCell(m_ordFunction));
  m_defaultValue = defaultValue;
  m_defaultValue = eval_value(m_defaultValue);

  eval(initList);
  while (whnfHead==nameCons) {
    if (whnfArgs!=2) internal("assoc list expected");
    eval(pop());
    if (!isTuple(whnfHead) || (whnfArgs!=2)) internal("tuple expected");
    arg   = pop();
    arg   = strict_eval(arg);
    value = pop();
    next  = pop();
    value = eval_value(value);
    writeSafeLocation((*m_values)[arg], Value(0,value), arg, true);
    m_initList = next;
    eval(next);
  } 
  m_initList = NIL;
  m_useCounter--;
  m_globals.initCounter--;
}

DynamicFunction::~DynamicFunction() {
#ifdef DEBUG_ASM
  printf("deleting dynamic function ");
  ERREXPR(m_funname);
  printf(", id = %d\n",m_id);
#endif
  if (m_values) delete m_values;
}

int DynamicFunction::getNewID() {
  Globals_t &g = getGlobals();
  if (g.free==0) {
    if (g.flist==NULL) {
      g.size  = 10;
      g.free  = g.size;
      g.flist = new (DynamicFunction*)[g.size];
      for (int i=0;i<g.size;i++) g.flist[i] = NULL;
      g.free--;
      return 0;
    }
    else {
      DynamicFunction **old = g.flist;
      g.flist = new (DynamicFunction*)[2*g.size];
      for (int i=0;i<g.size;i++) {
	g.flist[i]        = old[i];
	g.flist[i+g.size] = NULL;
      }
      g.free = g.size;
      g.size = 2 * g.size;
      delete old;
      g.free--;
      return (g.free+1);
    }
  }

  for (int i=0;i<g.size;i++)
    if (g.flist[i]==NULL) {
      g.free--;
      return i;
    }
  assert(0);
  return 0;
}

DynamicFunction::Globals_t &DynamicFunction::getGlobals() {
  static Globals_t g = { NULL,
			 0, 
			 0, 
			 0,
                         0,
                         0,
			 false,
			 map<int,LocationList_t>(),
			 map<int,LocationList_t>(),
			 NULL,
			 NULL
  };
  return g;
}


DynamicFunction *DynamicFunction::check(Cell asmFun) {
  Int save = evaluate_asm;
  evaluate_asm = 0;
  eval(asmFun);
  evaluate_asm = save; 
  if (whatIs(whnfHead)!=ASMFUN) {
    ERROR(0) "dynamic function expected" ETHEN
    ERRTEXT  "\n*** expression : " ETHEN ERREXPR(asmFun);
    ERRTEXT  "\n"
    EEND;
  }
  return getDynFun(snd(whnfHead));
}

void DynamicFunction::write(Cell newValue, bool force) {
  AsmValue_t *value;
  Cell arg = NIL;
  if (isInput()) {
    ERROR(0) "update of input function not allowed" ETHEN
    ERRTEXT "\n*** dynamic function : " ETHEN ERREXPR(funname());
    ERRTEXT "\n"
    EEND; 
  }
  if (m_dynFun==Dynamic0) {
    if (!force && !consistent()) return;
    if (whnfArgs!=0) { 
      ERROR(0) "no argument expected" ETHEN
      ERRTEXT "\n*** dynamic function : " ETHEN ERREXPR(funname());
      ERRTEXT "\n"
      EEND; 
    }
    value = &m_value;
  }
  else {
    if (whnfArgs!=1) { 
      ERROR(0) "exactly one argument expected" ETHEN
      ERRTEXT "\n*** dynamic function : " ETHEN ERREXPR(funname());
      ERRTEXT "\n"
      EEND; 
    }
    arg = pop();
    arg = strict_eval(arg);
    if (!force && !consistent()) return;
    value = &(*m_values)[arg];
  }
  if (force) {
    ValueList_t &l = value->oldValue;
    if (l.empty()) l.push_front(Value(0, eval_value(newValue)));
    else l.front() = Value(0, eval_value(newValue));
  }
  else
    writeSafeLocation(*value,
		      Value(m_globals.level, eval_value(newValue)),
		      arg,
		      false);
}

void DynamicFunction::writeLocation(AsmValue_t *v,
				    ValueList_t &l,
				    Value newValue,
				    Cell  arg,
				    bool init) {
  l.push_front(newValue);
  if (init) registerLocationOld(v,arg);
  else registerLocationNew(v,arg);
}

void DynamicFunction::writeSafeLocation(AsmValue_t &v,
					Value newValue,
					Cell  arg,
					bool init) {
  ValueList_t &l = init ? v.oldValue : v.newValue;
  if ((m_dynFun==Dynamic1) && eq(newValue.value(),m_defaultValue))
    newValue.setValue(ASMDEFAULT);
  if (l.empty() ||
      l.front().level()<m_globals.level)
    writeLocation(&v, l, newValue, arg, init);
  else {
    // check for consistency
    Cell old = l.front().value();
    if (!eq(old,newValue.value())) {
      if ((m_globals.ignoreInconsistence>0) &&
	  (m_globals.level>0)) {
	m_globals.isConsistent &= false;
      }
      else {
	ERROR(0) "inconsistent update" ETHEN
	ERRTEXT "\n*** dynamic function : " ETHEN ERREXPR(funname());
	ERRTEXT "\n*** expression (new) : " ETHEN ERREXPR(newValue.value());
	ERRTEXT "\n*** expression (old) : " ETHEN ERREXPR(old);
	ERRTEXT "\n"
	EEND; 
      }
    }
  }
}

Cell DynamicFunction::getAssocs() {
  Values_t::iterator it;
  Cell res, elem;
  ValueList_t *l;
  assert(m_dynFun == Dynamic1);
  push(nameNil);
  res = nameNil;
  if (!asmHistoryActive || 
      (asmHistoryLevel==asmHistoryStop-asmHistoryStart)) {
    for (it = m_values->begin(); it != m_values->end();it++) {
      l = &(*it).second.oldValue;
      if (!l->empty() && l->front().value()!=ASMDEFAULT) { 
	elem = ap(ap(nameCons,
		     ap(ap(mkTuple(2),(*it).first), l->front().value())
		     )
		, nameNil);
	if (top()==nameNil) top() = elem;
	else snd(res) = elem;
	res = elem;
      }
    }
  }
  else {
    Cell v;
    for (it = m_values->begin(); it != m_values->end();it++) {
      l = &(*it).second.oldValue;
      v = readHistoryValue(*l);
      if (v!=ASMDEFAULT) { 
	elem = ap(ap(nameCons,
		     ap(ap(mkTuple(2),(*it).first), v)
		     )
		, nameNil);
	if (top()==nameNil) top() = elem;
	else snd(res) = elem;
	res = elem;
      }
    }
  }
  return pop();
}

bool DynamicFunction::nonEmptyDom() {
  Values_t::iterator it;
  ValueList_t *l;
  assert(m_dynFun == Dynamic1);

  if (!asmHistoryActive || 
      (asmHistoryLevel==asmHistoryStop-asmHistoryStart)) {
    for (it = m_values->begin(); it != m_values->end();it++) {
      l = &(*it).second.oldValue;
      if (!l->empty() && l->front().value()!=ASMDEFAULT) return true;
    }
  }
  else {
    Cell v;
    for (it = m_values->begin(); it != m_values->end();it++) {
      l = &(*it).second.oldValue;
      v = readHistoryValue(*l);
      if (v!=ASMDEFAULT) return true;
    }
  }
  return false;
}

bool DynamicFunction::inDom(Cell key) {
  Values_t::iterator it;
  // key = strict_eval(key);
  it = m_values->find(key);
  if (it == m_values->end()) return false;
  if (!asmHistoryActive || 
      (asmHistoryLevel==asmHistoryStop-asmHistoryStart))
    return (!(*it).second.oldValue.empty() &&
	    (*it).second.oldValue.front().value() != ASMDEFAULT);
  else
    return (readHistoryValue((*it).second.oldValue)!=ASMDEFAULT);
}

void DynamicFunction::outputValue(Update_t *u, bool takeNew) {
  if (u->m_dynFun->outputLoc()) {
    printf("\n");
    asmPrintString(stdout,u->m_dynFun->funname());
    if (!u->m_dynFun->isDynamic0()) {
      printf(" ");
      printExp(stdout,u->m_arg);
    }
    printf(" = ");
  }
  if (takeNew) asmPrintString(stdout,u->m_v->newValue.front().value());
  else  asmPrintString(stdout,u->m_v->oldValue.front().value());
}

bool DynamicFunction::updateState(bool makeOutput) {
  // for all modified locations: move value from old to new
  Globals_t  &g = getGlobals();
  Update_t   *u;
  AsmValue_t *v;
  bool changed = false;
  while (!g.locationsNew->empty()) {
    u = &g.locationsNew->front();
    v = u->m_v;
    changed |= !u->m_dynFun->isInput();
    if (u->m_dynFun->isOutput() && makeOutput) 
      outputValue(u,true);
    else {
      if (v->oldValue.empty() || 
	  v->oldValue.front().level() < g.level) {
	v->oldValue.push_front(v->newValue.front());
	if (g.level>0) { 
	  g.locationsOld->push_front(*u);
	  u->m_dynFun->addUse();
	}
      }
      else
	v->oldValue.front() = v->newValue.front();
    }
    v->newValue.pop_front();
    u->m_dynFun->removeUse();
    g.locationsNew->pop_front();
  }
  return changed;
}

void DynamicFunction::closeFrame(bool writeSafe) {
  Globals_t &g = getGlobals();
  Update_t  *u;
  Value     *v;
  LocationList_t *g_new, *g_old;

  assert(g.level>0);
  g_new = g.locationsNew;
  g_old = g.locationsOld;
  
  if (!consistent()) {
    clearLevel(g.level>0);
    g.level--;
    g.locationsOld = &g.locationMapOld[g.level];
    g.locationsNew = &g.locationMapNew[g.level];  
  }
  else {
    assert(g_new->empty());
    g.level--;
    g.locationsOld = &g.locationMapOld[g.level];
    g.locationsNew = &g.locationMapNew[g.level];  
    while (!g_old->empty()) {
      u = &g_old->front();
      v = &u->m_v->oldValue.front();
      if (writeSafe) 
	u->m_dynFun->writeSafeLocation(*u->m_v, 
				       Value(g.level,v->value()),
				       u->m_arg,false);
      else {
	//	if (u->m_dynFun->isOutput()) outputValue(u,false);
	//	else 
	if (u->m_v->newValue.empty() ||
	    u->m_v->newValue.front().level() < g.level)
	  u->m_dynFun->writeLocation(u->m_v, 
				     u->m_v->newValue,
				     Value(g.level,v->value()),
				     u->m_arg,false);
      }
      u->m_dynFun->removeUse();
      u->m_v->oldValue.pop_front();
      g_old->pop_front();
    }
  }
}

void DynamicFunction::closeSecureFrame() {
  Globals_t  &g = getGlobals();
  Update_t   *u;
  AsmValue_t *v;
  LocationList_t *g_old;
  LocationList_t::iterator it;
  int newLevel = g.level-1;

  assert(g.locationsNew->empty());
  g_old = g.locationsOld;
  for (it = g_old->begin(); consistent() && (it != g_old->end()); it++) {
    u = &(*it);
    v = u->m_v;
    if (v->newValue.empty() ||
	(v->newValue.front().level() < newLevel) ||
        (u->m_dynFun->eq(v->newValue.front().value(), 
			 v->oldValue.front().value()))) ;
    else g.isConsistent = false;
  }

  closeFrame(false);

  g.ignoreInconsistence--;
  g.isConsistent = true;
}


void DynamicFunction::clearLevel(bool clearOld) {
  Globals_t &g = getGlobals();
  Update_t  *u;
  LocationList_t::iterator it;
#ifdef DEBUG_ASM
  if (!g.locationsNew->empty()) 
    printf("deleting temp values in level %d\n",g.level);
#endif
  for (it = g.locationsNew->begin(); it != g.locationsNew->end(); it++) {
    u = &(*it);
    u->m_v->newValue.pop_front();
    u->m_dynFun->removeUse();
  }
  g.locationsNew->clear();
  if (clearOld) {
  if (!g.locationsOld->empty()) 
#ifdef DEBUG_ASM
    printf("deleting values in level %d\n",g.level);
#endif
    for (it = g.locationsOld->begin(); it != g.locationsOld->end(); it++) {
      u = &(*it);
      u->m_v->oldValue.pop_front();
      u->m_dynFun->removeUse();
    }
    g.locationsOld->clear();
  }
}

void DynamicFunction::resetState() {
  Globals_t &g = getGlobals();
  g.level++;
  if (asmHistoryActive) {
    while (g.level>=asmHistoryStop-asmHistoryStart) {
      g.locationsNew = &g.locationMapNew[g.level];
      g.locationsOld = &g.locationMapOld[g.level];
      clearLevel(g.level>asmHistoryStop-asmHistoryStart);
      g.level--;
    }
    g.level = asmHistoryStop-asmHistoryStart;
    if (g.level<0) g.level = 0;
    g.ignoreInconsistence = false;
    g.isConsistent        = true;
  }
  else {
    do {
      g.locationsNew = &g.locationMapNew[g.level];
      g.locationsOld = &g.locationMapOld[g.level];
      clearLevel(g.level>0);
      g.level--;
    } while (g.level>=0);
    g.level = 0;
    g.ignoreInconsistence = false;
    g.isConsistent        = true;
  }
}

void DynamicFunction::clearAll() {
  Globals_t &g = getGlobals();
  asmHistoryStop  = 0;
  asmHistoryStart = 1;
  asmHistoryLevel = 0;
  resetState();
  asmHistoryStart = 0;
  used_steps      = 0;
  asmNewInteger   = 0;
  g.initCounter   = 0;
}

bool DynamicFunction::eq(Cell a, Cell b) {
  if (a==ASMDEFAULT) return (a==b);
  if (b==ASMDEFAULT) return (a==b);
  eval(ap(ap(m_eqFunction,a),b));
  return (whnfHead == nameTrue);
}


void DynamicFunction::registerLocationNew(AsmValue_t *v, Cell arg) {
  addUse();
  m_globals.locationsNew->push_front(Update_t(this,v,arg));
}

void DynamicFunction::registerLocationOld(AsmValue_t *v, Cell arg) {
  if (m_globals.level==0) return;
  addUse();
  m_globals.locationsOld->push_front(Update_t(this,v,arg));
}


void DynamicFunction::initGarbageCollection() {
  Globals_t &g = getGlobals();
  for (int i=0;i<g.size;i++)
    if (g.flist[i]) g.flist[i]->resetMark();
  for (int i=0;i<g.size;i++)
    if (g.flist[i] && g.flist[i]->used()) g.flist[i]->setMark();
}


void DynamicFunction::markCells() {
  Values_t::iterator it;
  Cell c;
  mark(m_funname);
  mark(m_eqFunction);
  mark(m_ordFunction);
  mark(m_defaultValue);
  mark(m_initList);
  if (m_values)
    for (it = m_values->begin(); it != m_values->end(); it++) {
      c = (*it).first;
      mark(c);
#ifdef DEBUG_ASM
      assert(c==(*it).first);
#endif
      markLocation((*it).second);
    }
  markLocation(m_value);
}

void DynamicFunction::markLocation(AsmValue_t &value) {
  ValueList_t::iterator it;
  Cell c;
  for (it = value.newValue.begin(); it != value.newValue.end(); it++) {
    c = (*it).value();
    mark(c);
#ifdef DEBUG_ASM
    assert(c==(*it).value());
#endif
  }
  for (it = value.oldValue.begin(); it != value.oldValue.end(); it++) {
    c = (*it).value();
    mark(c);
#ifdef DEBUG_ASM
    assert(c==(*it).value());
#endif
  }
}

void DynamicFunction::removeUnusedFunctions() {
 Globals_t &g = getGlobals();
 mark(last_evaluated_asmvar);
 for (int i=0;i<g.size;i++)
   if (g.flist[i])
     if (g.flist[i]->isMarked())
       g.flist[i]->markCells();
     else {
       delete g.flist[i];
       g.free++;
       g.flist[i] = NULL;
     }
}
